from axelrod import Player, Actions

C, D = Actions.C, Actions.D

class GoByMajority(Player):
    """A player examines the history of the opponent: if the opponent has more
    defections than cooperations then the player defects.

    An optional memory attribute will limit the number of turns remembered (by
    default this is 0)
    """

    classifier = {
        'stochastic': False,
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False,
        'memory_depth': 0 # memory_depth may be altered by __init__
    }

    def __init__(self, memory_depth=0, soft=True):
        """
        Parameters
        ----------
        memory_depth, int >= 0
            The number of rounds to use for the calculation of the cooperation
            and defection probabilities of the opponent.
        soft, bool
            Indicates whether to cooperate or not in the case that the
            cooperation and defection probabilities are equal.
        """

        Player.__init__(self)
        self.soft = soft
        self.classifier['memory_depth'] = memory_depth
        self.init_args = (memory_depth, soft)

    def strategy(self, opponent):
        """This is affected by the history of the opponent.

        As long as the opponent cooperates at least as often as they defect then the player will cooperate.
        If at any point the opponent has more defections than cooperations in memory the player defects.
        """

        memory = self.classifier['memory_depth']
        history = opponent.history[-memory:]
        defections = sum([s == D for s in history])
        cooperations = sum([s == C for s in history])
        if defections > cooperations:
            return D
        if defections == cooperations:
            if self.soft:
                return C
            else:
                return D
        return C

    def __repr__(self):
        """The string method for the strategy."""
        memory = self.classifier['memory_depth']
        name = 'Go By Majority' + (memory > 0) * (": %i" % memory)
        if self.soft:
            name = "Soft " + name
        else:
            name = "Hard " + name
        return name


class GoByMajority40(GoByMajority):
    """
    GoByMajority player with a memory of 40.
    """

    def __init__(self, memory_depth=40, soft=True):
        super(GoByMajority40, self).__init__(memory_depth=memory_depth,
                                             soft=soft)


class GoByMajority20(GoByMajority):
    """
    GoByMajority player with a memory of 20.
    """

    def __init__(self, memory_depth=20, soft=True):
        super(GoByMajority20, self).__init__(memory_depth=memory_depth,
                                             soft=soft)

class GoByMajority10(GoByMajority):
    """
    GoByMajority player with a memory of 10.
    """

    def __init__(self, memory_depth=10, soft=True):
        super(GoByMajority10, self).__init__(memory_depth=memory_depth,
                                             soft=soft)

class GoByMajority5(GoByMajority):
    """
    GoByMajority player with a memory of 5.
    """

    def __init__(self, memory_depth=5, soft=True):
        super(GoByMajority5, self).__init__(memory_depth=memory_depth,
                                             soft=soft)



from math import sqrt
from . import eigen
from axelrod import Actions

C, D = Actions.C, Actions.D


# As yet unused until RoundRobin returns interactions
def cooperation_matrix(interactions):
    """
    The cooperation matrix from a single round robin.

    Parameters
    ----------
    interactions : dictionary
        A dictionary of the form:

        e.g. for a round robin between Cooperator, Defector and Alternator
        with 2 turns per round:
        {
            (0, 0): [(C, C), (C, C)].
            (0, 1): [(C, D), (C, D)],
            (0, 2): [(C, C), (C, D)],
            (1, 1): [(D, D), (D, D)],
            (1, 2): [(D, C), (D, D)],
            (2, 2): [(C, C), (D, D)]
        }

        i.e. the key is a pair of player index numbers and the value, a list of
        plays. The list contains one pair per turn in the round robin.
        The dictionary contains one entry for each combination of players.

    nplayers : integer
        The number of players in the round robin

    Returns
    -------
    list
        The cooperation matrix (C) of the form:

            [
                [a, b, c],
                [d, e, f],
                [g, h, i],
            ]

        i.e. an n by n matrix where n is the number of players. Each row (i)
        and column (j) represents an individual player and the the value Cij
        is the number of times player i cooperated against opponent j.
    """

    # The number of ways (c) to select groups of r members from a set of n
    # members is given by:
    #
    #     c = n! / r!(n - r)!
    #
    # In this case, we are selecting pairs of players (p) and thus r = 2,
    # giving:
    #
    #     p = n(n-1) / 2 or p = (n^2 - n) / 2
    #
    # However, we also have the case where each player plays itself gving:
    #
    #     p = (n^2 + n) / 2
    #
    # Using the quadratic equation to rearrange for n gives:
    #
    #     n = (-1 +- sqrt(1 + 8p)) / 2
    #
    # Taking only the real roots allows us to derive the number of players
    # given the number of pairs:
    #
    #     n = (sqrt(8p + 1) -1) / 2

    nplayers = int((sqrt(len(interactions) * 8 + 1) - 1) / 2)
    cooperation = [[0 for i in range(nplayers)] for j in range(nplayers)]
    for players, actions in interactions.items():
        p1_actions, p2_actions = zip(*actions)
        p1_cooperation = p1_actions.count(C)
        p2_cooperation = p2_actions.count(C)
        cooperation[players[0]][players[1]] = p1_cooperation
        if players[0] != players[1]:
            cooperation[players[1]][players[0]] = p2_cooperation
    return cooperation


def cooperation(results):
    """
    The total cooperation matrix from a tournament of multiple repetitions.

    Parameters
    ----------
    results : list
        A matrix of the form:

        [
            [[a, j], [b, k], [c, l]],
            [[d, m], [e, n], [f, o]],
            [[g, p], [h, q], [i, r]],
        ]

        i.e. one row per player, containing one element per opponent (in
        order of player index) which lists cooperation values for each
        repetition.

    Returns
    -------
    list
        The cooperation matrix (C) of the form:

            [
                [[a + j], [b + k], [c + l]],
                [[d + m], [e + n], [f + o]],
                [[g + p], [h + q], [i + r]],
            ]

        i.e. an n by n matrix where n is the number of players. Each row (i)
        and column (j) represents an individual player and the the value Cij
        is the number of times player i cooperated against opponent j.
    """
    return[[sum(element) for element in row] for row in results]


def normalised_cooperation(cooperation, turns, repetitions):
    """
    The per-turn normalised cooperation matrix for a tournament of n repetitions.

    Parameters
    ----------
    cooperation : list
        The cooperation matrix (C)
    turns : integer
        The number of turns in each round robin.
    repetitions : integer
        The number of repetitions in the tournament.

    Returns
    -------
    list
        A matrix (N) such that:

            N = C / t

        where t is the total number of turns played in the tournament.
    """
    turns = turns * repetitions
    return[
        [1.0 * element / turns for element in row]
        for row in cooperation]


def vengeful_cooperation(cooperation):
    """
    The vengeful cooperation matrix derived from the cooperation matrix.

    Parameters
    ----------
    cooperation : list
        A cooperation matrix (C)

    Returns
    -------
    list
        A matrix (D) such that:

            Dij = 2(Cij -0.5)
    """
    return [[2 * (element - 0.5) for element in row] for row in cooperation]


def cooperating_rating(cooperation, nplayers, turns, repetitions):
    """
    A list of cooperation ratings for each player

    Parameters
    ----------
    cooperation : list
        The cooperation matrix
    nplayers : integer
        The number of players in the tournament.
    turns : integer
        The number of turns in each round robin.
    repetitions : integer
        The number of repetitions in the tournament.

    Returns
    -------
    list
        a list of cooperation rates ordered by player index
    """
    total_turns = turns * repetitions * nplayers
    return [1.0 * sum(row) / total_turns for row in cooperation]


def null_matrix(nplayers):
    """
    A null n by n matrix for n players

    Parameters
    ----------
    nplayers : integer
        The number of players in the tournament.
    Returns
    -------
    list
        A null n by n matrix where n is the number of players.
    """
    plist = list(range(nplayers))
    return [[0 for j in plist] for i in plist]


def good_partner_matrix(results, nplayers, repetitions):
    """
    An n by n matrix of good partner ratings for n players

    Parameters
    ----------
    results : list
        A cooperation results matrix of the form:

        [
            [[a, j], [b, k], [c, l]],
            [[d, m], [e, n], [f, o]],
            [[g, p], [h, q], [i, r]],
        ]

        i.e. one row per player, containing one element per opponent (in
        order of player index) which lists cooperation values for each
        repetition.

    nplayers : integer
        The number of players in the tournament.
    repetitions : integer
        The number of repetitions in the tournament.

    Returns
    -------
    list
        The good partner matrix (P) of the form:

        [
            [0, 0 + (1 if b >= d) + (1 if k >= m), 0 + (1 if c >= g) + (1 if l >= p) ],
            [0 + (1 if e >= g) + (1 if n >= p), 0, 0 + (1 if f >= h) + (1 if o >= q)],
            [0 + (1 if g >= c) + (1 if p >= l), 0 + (1 if h >= f) + (1 if q >= o), 0]
        ]

        i.e. an n by n matrix where n is the number of players. Each row (i)
        and column (j) represents an individual player and the the value Pij
        is the sum of the number of repetitions where player i cooperated as
        often or more than opponent j.
    """
    matrix = null_matrix(nplayers)
    for r in range(repetitions):
        for i in range(nplayers):
            for j in range(nplayers):
                if i != j and results[i][j][r] >= results[j][i][r]:
                    matrix[i][j] += 1
    return matrix


def n_interactions(nplayers, repetitions):
    """
    The number of interactions between n players

    Parameters
    ----------
    nplayers : integer
        The number of players in the tournament.
    repetitions : integer
        The number of repetitions in the tournament.

    Returns
    -------
    integer
        The number of interactions between players excluding self-interactions.
    """
    return repetitions * (nplayers - 1)


def good_partner_rating(good_partner_matrix, nplayers, repetitions):
    """
    A list of good partner ratings for n players in order of rating

    Parameters
    ----------
    good_partner_matrix : list
        The good partner matrix
    nplayers : integer
        The number of players in the tournament.
    repetitions : integer
        The number of repetitions in the tournament.

    Returns
    -------
    list
        A list of good partner ratings ordered by player index.
    """
    return [1.0 * sum(row) / n_interactions(nplayers, repetitions)
            for row in good_partner_matrix]


def eigenvector(cooperation_matrix):
    """
    The principal eigenvector of the cooperation matrix

    Parameters
    ----------
    cooperation_matrix : list
        A cooperation matrix

    Returns
    -------
    list
        The principal eigenvector of the cooperation matrix.
    """
    eigenvector, eigenvalue = eigen.principal_eigenvector(
        cooperation_matrix, 1000, 1e-3
    )
    return eigenvector.tolist()



from axelrod import Player, Actions, random_choice

"""IPD Strategies: http://www.prisoners-dilemma.com/strategies.html"""

C, D = Actions.C, Actions.D

class MemoryOnePlayer(Player):
    """Uses a four-vector for strategies based on the last round of play,
    (P(C|CC), P(C|CD), P(C|DC), P(C|DD)), defaults to Win-Stay Lose-Shift.
    Intended to be used as an abstract base class or to at least be supplied
    with a initializing four_vector."""

    name = 'Generic Memory One Player'
    classifier = {
        'memory_depth': 1,  # Memory-one Four-Vector
        'stochastic': True,
        'inspects_source': False,
        'manipulates_source': False,
        'manipulates_state': False
    }

    def __init__(self, four_vector=None, initial=C):
        """
        Parameters
        ----------
        fourvector, list or tuple of floats of length 4
            The response probabilities to the preceeding round of play
            ( P(C|CC), P(C|CD), P(C|DC), P(C|DD) )
        initial, C or D
            The initial move

        Special Cases
        -------------
        Alternator is equivalent to MemoryOnePlayer((0, 0, 1, 1), C)
        Cooperator is equivalent to MemoryOnePlayer((1, 1, 1, 1), C)
        Defector   is equivalent to MemoryOnePlayer((0, 0, 0, 0), C)
        Random     is equivalent to MemoryOnePlayer((0.5, 0.5, 0.5, 0.5))
           (with a random choice for the initial state)
        TitForTat  is equivalent to MemoryOnePlayer((1, 0, 1, 0), C)
        WinStayLoseShift is equivalent to MemoryOnePlayer((1, 0, 0, 1), C)
        See also: The remaining strategies in this file
                  Multiple strategies in titfortat.py
                  Grofman, Joss in axelrod_tournaments.py
        """
        Player.__init__(self)
        self._initial = initial
        if four_vector:
            self.set_four_vector(four_vector)
        self.init_args = (four_vector, initial)

    def set_four_vector(self, four_vector):
        if not all(0 <= p <= 1 for p in four_vector):
            raise ValueError('An element in the probability vector, %s, is not between 0 and 1.' % str(four_vector))

        self._four_vector = dict(zip([(C, C), (C, D), (D, C), (D, D)], map(float, four_vector)))
        self.classifier['stochastic'] = any(0 < x < 1 for x in set(four_vector))

    def strategy(self, opponent):
        if not hasattr(self, "_four_vector"):
            raise ValueError("_four_vector not yet set")
        if len(opponent.history) == 0:
            return self._initial
        # Determine which probability to use
        p = self._four_vector[(self.history[-1], opponent.history[-1])]
        # Draw a random number in [0,1] to decide
        return random_choice(p)


class WinStayLoseShift(MemoryOnePlayer):
    """Win-Stay Lose-Shift, also called Pavlov."""

    name = 'Win-Stay Lose-Shift'

    def __init__(self, initial=C):
        Player.__init__(self)
        self.set_four_vector([1,0,0,1])
        self._initial = initial
        self.init_args = (initial,)


class GTFT(MemoryOnePlayer):
    """Generous Tit-For-Tat Strategy."""

    name = 'GTFT'

    def __init__(self, p=None):
        """
        Parameters
        ----------
        p, float
            A parameter used to compute the four-vector

        Special Cases
        -------------
        TitForTat is equivalent to GTFT(0)
        """
        self.p = p
        super(GTFT, self).__init__()
        self.init_args = (p,)

    def receive_tournament_attributes(self):
        (R, P, S, T) = self.tournament_attributes["game"].RPST()
        if self.p is None:
            self.p = min(1 - float(T - R) / (R - S), float(R - P) / (T - P))
        four_vector = [1, self.p, 1, self.p]
        self.set_four_vector(four_vector)

    def __repr__(self):
        return "%s: %s" % (self.name, round(self.p, 2))


class StochasticCooperator(MemoryOnePlayer):
    """Stochastic Cooperator, http://www.nature.com/ncomms/2013/130801/ncomms3193/full/ncomms3193.html."""

    name = 'Stochastic Cooperator'

    def __init__(self):
        four_vector = (0.935, 0.229, 0.266, 0.42)
        super(self.__class__, self).__init__(four_vector)
        self.init_args = ()
        self.set_four_vector(four_vector)


class StochasticWSLS(MemoryOnePlayer):
    """Stochastic WSLS, similar to Generous TFT"""

    name = 'Stochastic WSLS'

    def __init__(self, ep=0.05):
        """
        Parameters
        ----------
        ep, float
            A parameter used to compute the four-vector -- the probability of
            cooperating when the previous round was CD or DC

        Special Cases
        -------------
        WinStayLoseShift is equivalent to StochasticWSLS(0)
        """

        self.ep = ep
        four_vector = (1.-ep, ep, ep, 1.-ep)
        super(StochasticWSLS, self).__init__(four_vector)
        self.init_args = (ep,)
        self.set_four_vector(four_vector)


class ZeroDeterminantPlayer(MemoryOnePlayer):
    """Abstraction for ZD players. The correct formula is Equation 14 in
    http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0077886 .
    These players enforce a linear difference in stationary payoffs
    s * (S_xy - l) = S_yx - l, yielding extortionate strategies with l = P and
    generous strategies when l = R and s > 0"""

    name = 'ZD ABC'

    def receive_tournament_attributes(self, phi=0, s=None, l=None):
        """
        Parameters
        ----------
        phi, s, l: floats
            Parameter used to compute the four-vector according to the
            parameterization of Zero Determinant strategies below.
        """

        (R, P, S, T) = self.tournament_attributes["game"].RPST()
        if s is None:
            s = 1
        if l is None:
            l = R

        # Check parameters
        s_min = - min((T-l) / (l-S), (l-S) / (T-l))
        if (l < P) or (l > R) or (s > 1) or (s < s_min):
            raise ValueError

        p1 = 1 - phi * (1 - s) * (R - l)
        p2 = 1 - phi * (s * (l - S) + (T - l))
        p3 = phi * ((l - S) + s * (T - l))
        p4 = phi * (1 - s) * (l - P)

        four_vector = [p1, p2, p3, p4]
        self.set_four_vector(four_vector)


class ZDGTFT2(ZeroDeterminantPlayer):
    """A Generous Zero Determinant Strategy."""

    name = 'ZD-GTFT-2'

    def __init__(self, phi=0.25, s=0.5):
        """
        Parameters
        ----------
        phi, s: floats
            Parameters passed through to ZeroDeterminantPlayer to determine
            the four vector.
        """
        self.phi = phi
        self.s = s
        super(ZDGTFT2, self).__init__()
        self.init_args = (phi, s)

    def receive_tournament_attributes(self):
        (R, P, S, T) = self.tournament_attributes["game"].RPST()
        self.l = R
        super(ZDGTFT2, self).receive_tournament_attributes(self.phi,
                                                                  self.s,
                                                                  self.l)

class ZDExtort2(ZeroDeterminantPlayer):
    """An Extortionate Zero Determinant Strategy."""

    name = 'ZD-Extort-2'

    def __init__(self, phi=1./9, s=0.5):
        """
        Parameters
        ----------
        phi, s: floats
            Parameters passed through to ZeroDeterminantPlayer to determine
            the four vector.
        """
        self.phi = phi
        self.s = s
        super(ZDExtort2, self).__init__()
        self.init_args = (phi, s)

    def receive_tournament_attributes(self):
        (R, P, S, T) = self.tournament_attributes["game"].RPST()
        self.l = P
        super(ZDExtort2, self).receive_tournament_attributes(self.phi,
                                                                  self.s,
                                                                  self.l)


### Strategies for recreating tournaments
# See also Joss in axelrod_tournaments.py

class SoftJoss(MemoryOnePlayer):
    """
    Defects with probability 0.9 when the opponent defects, otherwise
    emulates Tit-For-Tat.
    """

    name = "Soft Joss"

    def __init__(self, q=0.9):
        """
        Parameters
        ----------
        q, float
            A parameter used to compute the four-vector

        Special Cases
        -------------
        Cooperator is equivalent to SoftJoss(0)
        TitForTat  is equivalent to SoftJoss(1)
        """
        self.q = q
        four_vector = (1., 1 - q, 1, 1 - q)
        super(SoftJoss, self).__init__(four_vector)
        self.init_args = (q,)

    def __repr__(self):
        return "%s: %s" % (self.name, round(self.q, 2))



from __future__ import absolute_import
from collections import OrderedDict

import axelrod.strategies

from .tournament_manager import *


class TournamentManagerFactory(object):

    @classmethod
    def create_tournament_manager(
            cls,
            output_directory,
            no_ecological,
            rebuild_cache,
            cache_file,
            exclusions,
            processes,
            turns,
            repetitions,
            noise):

        kwargs = {
            'processes': processes,
            'turns': turns,
            'repetitions': repetitions,
            'noise': noise
        }

        manager = axelrod.TournamentManager(
            output_directory=output_directory,
            with_ecological=not no_ecological,
            save_cache=rebuild_cache,
            cache_file=cache_file)

        cls._add_tournaments(manager, exclusions, kwargs)

        return manager

    @classmethod
    def _add_tournaments(cls, manager, exclusions, kwargs):
        for name, strategies in cls._tournaments_dict(exclusions).items():
            players = manager.one_player_per_strategy(strategies)
            manager.add_tournament(
                name=name, players=players, **kwargs)

    @staticmethod
    def _tournaments_dict(exclusions=None):
        if exclusions is None:
            exclusions = []

        tournaments = OrderedDict([
            ('basic_strategies', axelrod.basic_strategies),
            ('strategies',
                axelrod.ordinary_strategies),
            ('cheating_strategies', axelrod.cheating_strategies),
            ('all_strategies',
                axelrod.ordinary_strategies +
                axelrod.cheating_strategies)
        ])

        return OrderedDict([
            (key, value) for
            key, value in tournaments.items() if key not in exclusions
        ])




from numpy.linalg import LinAlgError
from numpy import arange, mean, median
from warnings import warn

matplotlib_installed = True
try:
    import matplotlib.pyplot as plt
    import matplotlib.transforms as transforms
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    matplotlib_installed = False


class Plot(object):

    def __init__(self, result_set):
        self.result_set = result_set
        self.matplotlib_installed = matplotlib_installed

    ## Abstract Box and Violin plots

    def _boxplot(self, data, names, title=None):
        """For making boxplots."""
        if not self.matplotlib_installed:
            return None
        nplayers = self.result_set.nplayers
        width = max(nplayers / 3, 12)
        height = width / 2
        figure = plt.figure(figsize=(width, height))
        plt.boxplot(data)
        plt.xticks(self._boxplot_xticks_locations, names, rotation=90)
        plt.tick_params(axis='both', which='both', labelsize=8)
        if title:
            plt.title(title)
        return figure

    def _violinplot(self, data, names, title=None):
        """For making violinplots."""
        if not self.matplotlib_installed:
            return None
        nplayers = self.result_set.nplayers
        width = max(nplayers / 3, 12)
        height = width / 2
        figure = plt.figure(figsize=(width, height))
        spacing = 4
        positions = spacing * arange(1, nplayers + 1, 1)
        plt.violinplot(data, positions=positions, widths=spacing/2,
                       showmedians=True, showextrema=False)
        plt.xticks(positions, names, rotation=90)
        plt.xlim(0, spacing * (nplayers + 1))
        plt.tick_params(axis='both', which='both', labelsize=8)
        if title:
            plt.title(title)
        return figure

    ## Box and Violin plots for mean score, score diferrences, and wins

    @property
    def _boxplot_dataset(self):
        return [self.result_set.normalised_scores[ir] for ir in self.result_set.ranking]

    @property
    def _boxplot_xticks_locations(self):
        return list(range(1, len(self.result_set.ranked_names) + 2))

    @property
    def _boxplot_xticks_labels(self):
        return [str(n) for n in self.result_set.ranked_names]

    @property
    def _boxplot_title(self):
        return ("Mean score per stage game over {} "
                "turns repeated {} times ({} strategies)").format(
            self.result_set.turns,
            self.result_set.repetitions,
            len(self.result_set.ranking))

    def boxplot(self):
        """For the specific mean score boxplot."""
        data = self._boxplot_dataset
        names = self._boxplot_xticks_labels
        title = self._boxplot_title
        try:
            figure = self._violinplot(data, names, title=title)
        except LinAlgError:
            # Matplotlib doesn't handle single point distributions well
            # in violin plots. Should be fixed in next release:
            # https://github.com/matplotlib/matplotlib/pull/4816
            # Fall back to boxplot
            figure = self._boxplot(data, names, title=title)
        return figure

    @property
    def _winplot_dataset(self):
        # Sort wins by median
        wins = self.result_set.wins
        players = self.result_set.players
        medians = map(median, wins)
        medians = sorted([(m, i) for (i, m) in enumerate(medians)], reverse=True)
        # Reorder and grab names
        wins = [wins[x[-1]] for x in medians]
        ranked_names = [str(players[x[-1]]) for x in medians]
        return wins, ranked_names

    @property
    def _winplot_title(self):
        return ("Distributions of wins:"
                " {} turns repeated {} times ({} strategies)").format(
            self.result_set.turns,
            self.result_set.repetitions,
            len(self.result_set.ranking))

    def winplot(self):
        """Plots the distributions for the number of wins for each strategy."""
        if not self.matplotlib_installed:
            return None

        data, names = self._winplot_dataset
        title = self._winplot_title
        try:
            figure = self._violinplot(data, names, title)
        except LinAlgError:
            # Matplotlib doesn't handle single point distributions well
            # in violin plots. Should be fixed in next release:
            # https://github.com/matplotlib/matplotlib/pull/4816
            # Fall back to boxplot
            figure = self._boxplot(data, names, title)
        # Expand ylim a bit
        maximum = max(max(w) for w in data)
        plt.ylim(-0.5, 0.5 + maximum)
        return figure

    @property
    def _sdv_plot_title(self):
        return ("Distributions of payoff differences per stage game over {} "
                "turns repeated {} times ({} strategies)").format(
            self.result_set.turns,
            self.result_set.repetitions,
            len(self.result_set.ranking))

    @property
    def _sd_ordering(self):
        return self.result_set.ranking

        ## Sort by median then max
        #from operator import itemgetter
        #diffs = self.result_set.score_diffs
        #to_sort = [(median(d), max(d), i) for (i, d) in enumerate(diffs)]
        #to_sort.sort(reverse=True, key=itemgetter(0, 1))
        #ordering = [x[-1] for x in to_sort]
        #return ordering

    @property
    def _sdv_plot_dataset(self):
        ordering = self._sd_ordering
        diffs = self.result_set.score_diffs
        players = self.result_set.players
        # Reorder and grab names
        diffs = [diffs[i] for i in ordering]
        ranked_names = [str(players[i]) for i in ordering]
        return diffs, ranked_names

    def sdvplot(self):
        """Score difference violinplots to visualize the distributions of how
        players attain their payoffs."""
        diffs, ranked_names = self._sdv_plot_dataset
        title = self._sdv_plot_title
        figure = self._violinplot(diffs, ranked_names, title)
        return figure

    ## Payoff heatmaps

    @property
    def _payoff_dataset(self):
        return [[self.result_set.payoff_matrix[r1][r2]
                for r2 in self.result_set.ranking]
                for r1 in self.result_set.ranking]

    @property
    def _pdplot_dataset(self):
        # Order like the sdv_plot
        ordering = self._sd_ordering
        pdm = self.result_set.payoff_diffs_means
        # Reorder and grab names
        matrix = [[pdm[r1][r2] for r2 in ordering]
                  for r1 in ordering]
        players = self.result_set.players
        ranked_names = [str(players[i]) for i in ordering]
        return matrix, ranked_names

    def _payoff_heatmap(self, data, names, title=None):
        """Generic heatmap plot"""
        if not self.matplotlib_installed:
            return None

        nplayers = self.result_set.nplayers
        width = max(nplayers / 4, 12)
        height = width
        figure, ax = plt.subplots()
        figure.set_figwidth(width)
        figure.set_figheight(height)
        mat = ax.matshow(data, cmap='YlGnBu')
        plt.xticks(range(self.result_set.nplayers))
        plt.yticks(range(self.result_set.nplayers))
        ax.set_xticklabels(names, rotation=90)
        ax.set_yticklabels(names)
        plt.tick_params(axis='both', which='both', labelsize=16)
        # Make the colorbar match up with the plot
        divider = make_axes_locatable(plt.gca())
        cax = divider.append_axes("right", "5%", pad="3%")
        plt.colorbar(mat, cax=cax)
        if title:
            plt.title(title)
        return figure

    def pdplot(self):
        """Payoff difference heatmap to visualize the distributions of how
        players attain their payoffs."""
        matrix, names = self._pdplot_dataset
        return self._payoff_heatmap(matrix, names)

    def payoff(self):
        """Payoff heatmap to visualize the distributions of how
        players attain their payoffs."""
        data = self._payoff_dataset
        names = self.result_set.ranked_names
        return self._payoff_heatmap(data, names)

    ## Ecological Plot

    def stackplot(self, eco):

        if not self.matplotlib_installed:
            return None

        if type(eco) is list:
            warn("""Passing the population sizes as an argument is deprecated and will be removed, please pass the Ecosystem directly""")
            populations = eco
        else:
            populations = eco.population_sizes

        figure, ax = plt.subplots()
        turns = range(len(populations))
        pops = [[populations[iturn][ir] for iturn in turns] for ir in self.result_set.ranking]
        ax.stackplot(turns, *pops)

        ax.yaxis.tick_left()
        ax.yaxis.set_label_position("right")
        ax.yaxis.labelpad = 25.0

        plt.ylim([0.0, 1.0])
        plt.ylabel('Relative population size')
        plt.xlabel('Turn')
        plt.title("Strategy population dynamics based on average payoffs")

        trans = transforms.blended_transform_factory(ax.transAxes, ax.transData)
        ticks = []
        for i, n in enumerate(self.result_set.ranked_names):
            x = -0.01
            y = (i + 0.5) * 1.0 / self.result_set.nplayers
            ax.annotate(n, xy=(x, y), xycoords=trans, clip_on=False, va='center', ha='right', fontsize=5)
            ticks.append(y)
        ax.set_yticks(ticks)
        ax.tick_params(direction='out')
        ax.set_yticklabels([])

        ax.set_xscale('log')

        return figure
